/***************************************************************************

Christer Karlsson

diningp2.c	-	main program that simulates a
				dining philosophers problem,
				This is not a dead-lock free version
				Argument N passed to diningp1 is
				used to create N philosophers.

				1)	Run diningp2 to simulate dining
					philosopher problem.
					The program accept one command line
					argument of integer type, this integer
					is used to create certain number of
					philosopers. This argument should be
					less than or equal to 20.

   					./diningp2 10

 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

//number of philosophers
const int arraySize = 20;
int philNum;
int Available;
int Need[20];
int Allocation[20];

void semaphore()
{
  key_t key;
  int semid;
  int i;

  union semun
  {
    int val;                    /* value for SETVAL */
    struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
    unsigned short int *array;  /* array for GETALL, SETALL */
    struct seminfo *__buf;      /* buffer for IPC_INFcontroller */
  } controller;

  /*
    semun's members:
    int val;
    struct semid_ds *buf;
    ushort *array;
  */

  if ( (key = ftok("seminit.c", 'J')) == -1 )
  {
    perror( "ftok" );
    exit( 1 );
  }

  /* create semaphore set with philNum semaphores */
  if ( (semid = semget(key, philNum, 0666 | IPC_CREAT)) == -1 )
  {
    perror( "semget" );
    exit( 1 );
  }

  /* initialize all semaphores in array to 1 */
  controller.val = 1;

  for ( i = 0; i < philNum; i++ )
  {
    if ( semctl(semid, i, SETVAL, controller) == -1 )
    {
      perror( "semctl" );
      exit( 1 );
    }
  }
}

//signal
void putFork( int philID, int forkID )
{
  key_t key;
  int semid;
  struct sembuf sb = {forkID, 1, 0};

  if ( (key = ftok("seminit.c", 'J')) == -1 )
  {
    perror( "ftok" );
    exit( 1 );
  }

  if ( (semid = semget(key, 1, 0)) == -1 )
  {
    perror( "semget" );
    exit( 1 );
  }

  if ( semop(semid, &sb, 1) == -1 )
  {
      perror( "semop" );
      exit( 1 );
  }
  Need[philID]++;
  Allocation[philID]--;
  Available++;
}

//wait
void takeFork( int philID, int forkID )
{
  key_t key;
  int semid;
  struct sembuf sb = {forkID, -1, 0};

  if ( (key = ftok("seminit.c", 'J')) == -1 )
  {
    perror( "ftok" );
    exit( 1 );
  }

  if ( (semid = semget(key, 1, 0)) == -1 )
  {
    perror( "semget" );
    exit( 1 );
  }

  if ( semop(semid, &sb, 1) == -1 )
  {
      perror( "semop" );
      exit( 1 );
  }
  Need[philID]--;
  Allocation[philID]++;
  Available--;
}

void think( int philID )
{
  printf( "[%i]\t: Thinking starts now....\n", philID + 1 );
  sleep( rand() % 5 );
  printf( "[%i]\t: Thinking over....\n", philID + 1 );
}

void eat( int philID )
{
  printf( "[%i]\t: Eating starts now....\n", philID + 1 );
  sleep( rand() % 5 );
  printf( "[%i]\t: Eating over....\n", philID + 1 );
}

int safety(k)
{
  int i=0;
  int safe=1;
  int Finish[philNum];
  int work=Available-1;

  // Set the ability to Finish to false
  // if any resources are allocated
  for(i=0;i<philNum;i++)
  {
	 if( Allocation[i]!=0)
	    Finish[i]=0;
     else
        Finish[i]=1;
  }

  i=-1;
  Finish[k]=0;

  // Check to see if all processes are able
  // to finish if the resorces is allocated
  while(i<philNum)
  {
     i++;
     if(Finish[i]==0 && work>=(Need[i]))
     {
        work=work+Allocation[i]+1;
        Finish[i]=1;
        i=-1;
     }

	 for(i=0;i<philNum;i++)
     {
        if(Finish[i]==0)
           safe=0;
     }

  }
  // Return 0 if it is unsafe to allocate
  // else Return 1
  return(safe);
}
int request(int i)
{
  int safe=0;
  // If there are still any resouces available
  // Check to see if it safe to hand them out
  if (Available>0)
	safe=safety(i);

  // Return 0 if request is denied
  // else Return 1
  return(safe);
}

void * philosopher( int i )
{
  printf( "[%i]\t: Philosopher #%d joined the dinner.\n", i + 1, i + 1 );

  while ( 1 )
  {
    think( i ); //philosopher is thinking

    printf( "[%d]\t: Trying to take left chopstick...\n", i + 1 );
	// Check to see if resource can be handed out
	if(request(i)==0);
    else
      takeFork( i, (i > 0) ? i - 1 : philNum - 1 ); //take left chopstick
    printf( "[%d]\t: Took left chopstick...\n", i + 1 );

    sleep( rand() % 5 ); //rests before taking second chopstick

    printf( "[%d]\t: Trying to take right chopstick...\n", i + 1 );
	// Check to see if resource can be handed out
	if(request(i)==0);
    else
	  takeFork( i, i ); //take right chopstick
    printf( "[%d]\t: Took right chopstick...\n", i + 1 );

    eat( i ); //philosopher eats

    printf( "[%d]\t: Puts left chopstick on the table...\n", i + 1 );
    putFork( i, (i > 0) ? i - 1 : philNum - 1 ); //put left chopstick the on table
    printf( "[%d]\t: Puts right chopstick on the table...\n", i + 1 );
    putFork( i, i ); //put right chopstick back on table
  }
}

int main( int argc, char ** argv )
{
  int i;
  int * status;
  philNum = atoi( argv[1] );

  if(philNum > 20 || philNum < 1)	//Make sure 0<philNum<21
    philNum=20;

  semaphore();						//Create the semphores # equals number of phil.
  pid_t pid;
  pthread_t philosophers[philNum];
  Available = philNum;
  srand(time(NULL));

  for (i=0; i < arraySize;i++)
  {
    Need[i]=2;
    Allocation[i]=0;
  }

  for ( i = 0; i < philNum; i++ )
  {
    pthread_create( &philosophers[i], NULL, (void *) philosopher, (void *) i );
  }

  for ( i = 0; i < philNum; i++ )
  {
    pthread_join( philosophers[i], NULL );
  }

  return 0;
}
